from pmc_automation_tools.api.ux.datasource import UXDataSource, UXDataSourceInput
from pmc_automation_tools.api.classic.datasource import ClassicDataSource, ClassicDataSourceInput
from pmc_automation_tools.api.datasource import ApiDataSource, ApiDataSourceInput
from pmc_automation_tools.common.utils import debug_logger, create_batch_folder, setup_logger, read_updated, save_updated
from pmc_automation_tools.driver.ux.driver import UXDriver
from pmc_automation_tools.driver.classic.driver import ClassicDriver
from pmc_automation_tools.driver.generic import GenericDriver
from pmc_automation_tools.driver.common import (
    VISIBLE,
    INVISIBLE,
    CLICKABLE,
    EXISTS
)

__version__ = "0.3.0"
__all__ = [
    "UXDataSource",
    "UXDataSourceInput",
    "ClassicDataSource",
    "ClassicDataSourceInput",
    "ApiDataSource",
    "ApiDataSourceInput",
    "debug_logger",
    "create_batch_folder",
    "setup_logger",
    "read_updated",
    "save_updated",
    "UXDriver",
    "ClassicDriver",
    "VISIBLE",
    "INVISIBLE",
    "CLICKABLE",
    "EXISTS",
    "GenericDriver"
]


from typing import Literal, Union
from pmc_automation_tools.driver.common import (
    PlexDriver,
    PlexElement,
    VISIBLE,
    INVISIBLE,
    CLICKABLE,
    EXISTS,
    )
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support.ui import Select



class GenericDriver(PlexDriver):
    def __init__(self, driver_type: Literal['edge', 'chrome'], *args, **kwargs):
        super().__init__(environment='ux', *args, driver_type=driver_type, **kwargs)
        for k, v in kwargs.items():
            setattr(self, k, v)
    
    def wait_for_element(self, selector, *args, driver=None, timeout=15, type=VISIBLE, ignore_exception=False) -> 'GenericElement':
        return super().wait_for_element(selector, *args, driver=driver, timeout=timeout, type=type, ignore_exception=ignore_exception, element_class=GenericElement)
    
    def login(self, url) -> 'GenericDriver':
        self.driver = self._driver_setup(self.driver_type)
        self.driver.get(url)
        return self.driver
    launch = login
    
    def click_button(self):...
    def token_get(self):...
    def _pcn_switch(self):...

class GenericElement(PlexElement):
    def __init__(self, webelement, parent):
        super().__init__(webelement, parent)
    
    def sync_picker(self, sync_value:Union[int, str], text:bool=True) -> None:
        if not isinstance(sync_value, (int, str)):
            raise TypeError("sync_value must be an int or str")
        _select = Select(self)
        if isinstance(sync_value, str):
            if text:
                _select.select_by_visible_text(sync_value)
                return None
            _select.select_by_value(sync_value)
            return None
        _select.select_by_index(sync_value)


from typing import Literal, Union
from pmc_automation_tools.driver.common import (
    PlexDriver,
    PlexElement,
    VISIBLE,
    INVISIBLE,
    CLICKABLE,
    EXISTS,
    SIGNON_URL_PARTS
    )
from selenium.common.exceptions import (
    TimeoutException,
    StaleElementReferenceException,
    NoSuchElementException
    )
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support.ui import Select
from pmc_automation_tools.common.exceptions import (
    UpdateError,
    NoRecordError,
    LoginError
    )
import time
BANNER_SUCCESS = 1
BANNER_WARNING = 2
BANNER_ERROR = 3
BANNER_CLASSES = {
    'plex-banner-success': BANNER_SUCCESS,
    'plex-banner-error': BANNER_WARNING,
    'plex-banner-warning': BANNER_ERROR
}
BANNER_SELECTOR = (By.CLASS_NAME, 'plex-banner')
PLEX_GEARS_SELECTOR = (By.XPATH, '//i[@class="plex-waiting-spinner"]')
UX_INVALID_PCN_MESSAGE = '__MESSAGE=YOU+WERE+REDIRECTED+TO+YOUR+LANDING+COMPANY'


class UXDriver(PlexDriver):
    def __init__(self, driver_type: Literal['edge', 'chrome'], *args, **kwargs):
        super().__init__(environment='ux', *args, driver_type=driver_type, **kwargs)
        for k, v in kwargs.items():
            setattr(self, k, v)


    def wait_for_element(self, selector, *args, driver:Union['UXDriver','UXPlexElement']=None, timeout=15, type=VISIBLE, ignore_exception=False) -> 'UXPlexElement':
        return super().wait_for_element(selector, *args, driver=driver, timeout=timeout, type=type, ignore_exception=ignore_exception, element_class=UXPlexElement)


    def wait_for_elements(self, selector, *args, driver:Union['UXDriver','UXPlexElement']=None, timeout=15, type=VISIBLE, ignore_exception=False) -> 'UXPlexElement':
        return super().wait_for_elements(selector, *args, driver=driver, timeout=timeout, type=type, ignore_exception=ignore_exception, element_class=UXPlexElement)


    def wait_for_banner(self, timeout:int=10, ignore_exception:bool=False) -> None:
        """Wait for the banner to appear and handle success/error/warning messages.

        Args:
            timeout (int, optional): How long to wait for the banner before throwing an exception. Defaults to 10.
            ignore_exception (bool, optional): Don't re-raise an exception if the banner is not found. Defaults to False.

        Raises:
            UpdateError: Unexpected banner type.
            UpdateError: No banner detected at all.
            UpdateError: Any non-success banners raise an error with the banner text.
        """        
        try:
            loop = 0
            while loop <= timeout:
                banner = self.wait_for_element(BANNER_SELECTOR)
                banner_class = banner.get_attribute('class')
                banner_type = next((BANNER_CLASSES[c] for c in BANNER_CLASSES if c in banner_class), None)
                if banner_type:
                    self._banner_handler(banner_type, banner)
                    break
                time.sleep(1)
                loop += 1
            else:
                if ignore_exception:
                    return None
                raise UpdateError(f'Unexpected banner type detected. Found {banner_class}. Expected one of {list(BANNER_CLASSES.keys())}')
        except (TimeoutException, NoSuchElementException, StaleElementReferenceException):
            if ignore_exception:
                return None
            raise UpdateError('No banner detected.')


    def _banner_handler(self, banner_type, banner):
        if banner_type == BANNER_SUCCESS:
            return 
        else:
            banner_text = banner.get_property('textContent')
            raise UpdateError(banner_text)
    

    def wait_for_gears(self, loading_timeout=10) -> None:
        super().wait_for_gears(PLEX_GEARS_SELECTOR, loading_timeout)

    def click_button(self, button_text:str, driver:Union['UXDriver','UXPlexElement']=None) -> None:
        """Clicks a standard button with matching text.

        Args:
            button_text (str): Text of the button to click.
            driver (UXDriver | UXPlexElement, optional): webdriver root to use if different than default. Defaults to None.

        Usage:

            If you don't provide the root driver, then the main page's Ok button will be clicked and not the popup window's button.
            ::

                popup_window = driver.find_element(By.ID, 'popupID')
                click_button('Ok', driver=popup_window)

            Alternatively:
            ::

                pa = UXDriver()
                popup_window = pa.wait_for_element((By.ID, 'popupID'))
                popup_window.click_button('Ok')
                
        """
        driver = driver or self.driver
        buttons = driver.find_elements(By.TAG_NAME, 'button')
        for b in buttons:
            if b.get_property('textContent') == button_text:
                self.debug_logger.debug(f'Button found with matching text: {button_text}')
                b.click()
                break
            
    def click_action_bar_item(self, item:str, sub_item:str=None) -> None:
        """Clicks on an action bar item.

        Args:
            item (str): Text for the item to click
            sub_item (str, optional): Text for the item if it is within a dropdown after clicking the item. Defaults to None.
        """
        action_bar = self.wait_for_element((By.CLASS_NAME, 'plex-actions'))

        # Check for the "More" link and determine if it's visible
        try:
            more_box = action_bar.find_element(By.LINK_TEXT, "More")
            style = more_box.find_element(By.XPATH, 'ancestor::li').get_dom_attribute('style')
            more_visible = 'none' not in style
            self.debug_logger.debug(f"'More' link {'found and visible' if more_visible else 'found but not visible'}.")
        except NoSuchElementException:
            self.debug_logger.debug('No element found for "More" link.')
            more_visible = False
        # Click on "More" if visible and adjust the action bar
        if more_visible:
            self.debug_logger.debug('Clicking "More" button.')
            more_box.click()
            self.wait_for_element((By.CLASS_NAME, "plex-subactions.open"))
            action_bar = self.wait_for_element((By.CLASS_NAME, 'plex-actions-more'))

        # Handle sub_item or main item click
        if sub_item:
            self.debug_logger.debug("Clicking sub-item.")
            self._click_sub_item(action_bar, item, sub_item)
        else:
            self.debug_logger.debug("Clicking main item.")
            action_item = self.wait_for_element((By.LINK_TEXT, item), type=CLICKABLE)
            action_item.click()

    def _click_sub_item(self, action_bar, item, sub_item):
        """Helper function to click on a sub-item."""
        action_items = action_bar.find_elements(By.CLASS_NAME, "plex-actions-has-more")
        for a in action_items:
            span_texts = a.find_elements(By.TAG_NAME, 'span')
            for s in span_texts:
                if s.get_property('textContent') == item:
                    s.find_element(By.XPATH, "ancestor::a").click()
                    break
        action_bar.find_element(By.LINK_TEXT, sub_item).click()


    def login(self, username, password, company_code, pcn, test_db=True, headless=False):
        self._set_login_vars()
        super().login(username, password, company_code, pcn, test_db, headless)
        self._login_validate()
        self.pcn_switch(self.pcn)
        self.token = self.token_get()
        self.first_login = False
        return (self.driver, self.url_comb, self.token)
    
    
    def _set_login_vars(self):
        self.plex_main = 'cloud.plex.com'
        self.plex_prod = ''
        self.plex_test = 'test.'
        self.sso = '/sso'
        super()._set_login_vars()


    def token_get(self) -> str:
        url = self.driver.current_url
        url_split = url.split('/')
        url_proto = url_split[0]
        url_domain = url_split[2]
        self.url_comb = f'{url_proto}//{url_domain}'
        self.url_token = url.split('?')[1]
        if '&' in self.url_token:
            self.url_token = [x for x in self.url_token.split('&') if 'asid' in x][0]
        return self.url_token
    
    
    def _pcn_switch(self, pcn=None):
        if not pcn:
            pcn = self.pcn
        if self.first_login:
            self.first_login = False
            return
        self.url_token = self.token_get()
        self.driver.get(f'{self.url_comb}/SignOn/Customer/{pcn}?{self.url_token}')
        if UX_INVALID_PCN_MESSAGE in self.driver.current_url.upper():
            raise LoginError(self.environment, self.db, pcn, f'Unable to login to PCN. Verify you have access.')
    
    
    def _login_validate(self):
        url = self.driver.current_url
        if not any(url_part in url.upper() for url_part in SIGNON_URL_PARTS):
            raise LoginError(self.environment, self.db, self.pcn_name, 'Login page not detected. Please validate login credentials and try again.')
        


class UXPlexElement(PlexElement):
    def __init__(self, webelement, parent):
        super().__init__(webelement, parent)


    def sync_picker(self, text_content:str, clear:bool=False, date:bool=False, column_delimiter:str='\t') -> None:
        """Sync the picker element to the provided value.

        Args:
            text_content (str): Desired value for the picker
            clear (bool, optional): Clear the picker if providing a blank text_content. Defaults to False.
            date (bool, optional): If the picker is a date picker. This should be detected automatically, but can be forced if behavior is unexpected.. Defaults to False.
            column_delimiter (str, optional): Delimiter for splitting the popup row's text for searching exact matches. Defaults to '\t'. 

        Raises:
            NoRecordError: If the drop-down picker does not have a matching option with the provided text content.
            NoRecordError: If the popup window does not return any results with the provided text content.
            NoSuchElementException: If the popup window has results, but there was no element matching the text content.
        """
        if not text_content and not clear:
            return
        picker_type = self.get_attribute('class')
        if picker_type == 'input-sm':
            date = True
        if self.tag_name == 'select':
            self.debug_logger.debug('Picker type is selection list.')
            self._handle_select_picker(text_content)
            return
        self.debug_logger.debug('Handling non-select picker.')
        # Check for existing selected item
        matching = self._check_existing_selection(text_content)

        # If there's no existing matching item, send the new text and process the picker
        if not matching:
            self.send_keys(text_content)
            self.send_keys(Keys.TAB)
            self._handle_popup_or_picker(text_content, date, column_delimiter)
    
    
    def _handle_select_picker(self, text_content):
        _select = Select(self)
        current_selection = _select.first_selected_option.text
        if current_selection == text_content:
            self.debug_logger.debug(f'Picker selection: {current_selection} matches {text_content}')
            return
        matching = any(o.text == text_content for o in _select.options)
        if matching:
            self.debug_logger.info(f'Matching option found. Picking {text_content}')
            _select.select_by_visible_text(text_content)
            self.send_keys(Keys.TAB)
        else:
            self.debug_logger.info(f'No matching selection available for {text_content}')
            raise NoRecordError(f'No matching selection available for {text_content}')
        

    def _check_existing_selection(self, text_content):
        try:
            self.debug_logger.debug('Trying to locate an existing selected item.')
            selected_element = self.wait_for_element(
                (By.XPATH, "preceding-sibling::div[@class='plex-picker-selected-items']"),
                driver=self,
                timeout=1
            )
            if selected_element:
                current_text = self.wait_for_element(
                    (By.CLASS_NAME, "plex-picker-item-text"),
                    driver=selected_element
                ).get_property('textContent')
                
                if current_text == text_content:
                    self.debug_logger.debug(f'Current text: {current_text} matches provided text: {text_content}.')
                    return True
                
                self.debug_logger.info(f'Current text: {current_text} does not match provided text: {text_content}')
                self.send_keys(Keys.BACKSPACE)
                self.clear()
        except (NoSuchElementException, TimeoutException):
            self.debug_logger.debug('No initial selected item detected.')
        return False
    

    def _handle_popup_or_picker(self, text_content, date, column_delimiter):
        try:
            picker_xpath = "preceding-sibling::div[@class='plex-picker-item']" if date else "preceding-sibling::div[@class='plex-picker-selected-items']"
            self.wait_for_element((By.XPATH, picker_xpath), driver=self, timeout=5)
            self.debug_logger.info(f'Picker has been filled in with {text_content}')
        except (TimeoutException, NoSuchElementException):
            self._handle_popup_window(text_content, column_delimiter)


    def _handle_popup_window(self, text_content, column_delimiter):
        try:
            self.debug_logger.debug('Checking for a popup window.')
            popup = self.wait_for_element((By.CLASS_NAME, 'modal-dialog.plex-picker'), timeout=3)
            multi = 'plex-picker-multi' in popup.get_attribute('class')
            self.wait_for_gears()

            items = popup.find_elements(By.CLASS_NAME, 'plex-grid-row')
            if not items:
                self._handle_no_records_popup(popup, text_content)
            
            option_found = self._find_and_click_option(items, text_content, column_delimiter)
            
            if not option_found:
                raise NoSuchElementException(f'No matching elements found for {text_content}')
            if multi:
                self.debug_logger.info('Multi-picker, clicking ok on the popup window.')
                self.click_button('Ok', driver=popup)

        except (TimeoutException, NoSuchElementException):
            self.debug_logger.info(f'No matching elements found for {text_content}')
            raise


    def _handle_no_records_popup(self, popup, text_content):
        result_text = popup.find_element(By.TAG_NAME, 'h4').get_property('textContent')
        if 'No records' in result_text:
            self.debug_logger.info(f'No records found for {text_content}')
            popup.find_element(By.CLASS_NAME, 'modal-footer').find_element(By.LINK_TEXT, 'Cancel').click()
            raise NoRecordError(f'No records found for {text_content}')

    def _find_and_click_option(self, items, text_content, column_delimiter):
        for item in items:
            item_columns = item.get_property('innerText').split(column_delimiter)
            if text_content in item_columns:
                self.debug_logger.info(f'Found matching item with text {item.text}.')
                item.click()
                return True
        return False



# UX Datasource
import os
import json
from warnings import warn
from requests.auth import HTTPBasicAuth
from pmc_automation_tools.api.common import (
    DataSourceInput,
    DataSourceResponse,
    DataSource,
    CustomSslContextHTTPAdapter,
    RETRY_COUNT,
    BACKOFF,
    RETRY_STATUSES
    )
from pmc_automation_tools.common.exceptions import(
    UXResponseErrorLog
)
import requests
from urllib3.util.retry import Retry
from itertools import chain
class UXDataSourceInput(DataSourceInput):
    def __init__(self, data_source_key: str, *args, template_folder: str=None, **kwargs):
        super().__init__(data_source_key, type='ux', *args, **kwargs)
        self.__input_types__ = {}
        self.__template_folder__ = template_folder
        if self.__template_folder__:
            template_query = self._query_template_import()
            if template_query:
                for key, value in template_query.items():
                    setattr(self, key, value)
        self._type_create()


    def _query_template_import(self):
        for file in os.listdir(self.__template_folder__):
            if f'{self.__api_id__}.json' == file:
                with open(os.path.join(self.__template_folder__, file), 'r', encoding='utf-8') as j:
                    template = json.loads(j.read())
                if 'inputs' in template.keys():
                    return template['inputs']
                return template


    def _update_input_parameters(self):
        self._query_string = {k:v for k, v in vars(self).items() if not k.startswith('_')}


    def _type_create(self):
        for k, v in vars(self).items():
            if not v or k.startswith('_'):
                continue
            value_type = type(v)
            if value_type is int and len(str(v)) == 1:
                self.__input_types__[k] = bool
            else:
                self.__input_types__[k] = value_type


    def get_type(self, attribute):
        """
        Return the expected input type for the provided attribute.
        
        Expects a data source template file to be used during object creation.

        Parameters:
        
        - attribute: Data source input name

        Returns:
        
        - type of the attribute derived from the data source template file.
        """
        return getattr(self, '__input_types__').get(attribute, None)

    def _xstr(self, s):
        return str(s or '')


    def _xbool(self, b):
        if isinstance(b, int):
            return bool(b)
        return b.strip().upper() != 'FALSE'


    def type_reconcile(self):
        """
        Adjusts the object attribute types to match the expected types of the data source.
        """
        for k, v in vars(self).items():
            if k.startswith('_') or k not in self.__input_types__.keys() or v is None:
                continue
            target_type = getattr(self, '__input_types__')[k]
            if target_type is int:
                new_val = None if isinstance(v, str) and not v.strip() else target_type(v)
            elif target_type is str:
                new_val = self._xstr(v)
            elif target_type is bool:
                new_val = self._xbool(v)
            else:
                new_val = target_type(v)
            setattr(self, k, new_val)


    def get_to_update(self, get_instance:'UXDataSourceResponse'):
        """
        Adjusts the attribute types to match the expected types of the data source.

        Parameters:
        - get_instance: a UXDataSourceResponse object returned from a call to a 'get' type datasource.
        """
        for k, v in vars(get_instance).items():
            setattr(self, k, v)
        for k in self.__input_types__.keys():
            if k not in vars(get_instance) and not k.startswith('_'):
                self.pop_inputs(k)
        self.type_reconcile()
        self.purge_empty()

    def purge_empty(self):
        """
        Removes empty/Nonetype attributes from the input.

        Additionally removes any attributes not existing in the input_types dictionary.
        """
        super().purge_empty()
        purge_attrs = []
        for y in vars(self).keys():
            if y not in self.__input_types__.keys() and not y.startswith('_'):
                purge_attrs.append(y)
        for y in purge_attrs:
            self.pop_inputs(y)

class UXDataSource(DataSource):
    def __init__(self, auth: HTTPBasicAuth | str,
                 *args,
                 test_db: bool = True,
                 pcn_config_file: str = 'resources/pcn_config.json',
                 **kwargs):
        """
        Parameters:

        - auth: HTTPBasicAuth | str
            - HTTPBasicAuth object
            - PCN Reference key for getting the username/password in a json config file.
            
        - test_db: bool, optional
            - Use test or production database
        
        - pcn_config_file: str, optional
            - Path to JSON file containing username/password credentials for HTTPBasicAuth connections.
        """
        super().__init__(*args, auth=auth, test_db=test_db, pcn_config_file=pcn_config_file, type='ux', **kwargs)
        self.url_db = 'test.' if self._test_db else ''
    
    def _create_session(self):
        session = requests.Session()
        retry = Retry(total=RETRY_COUNT, connect=RETRY_COUNT, backoff_factor=BACKOFF, status_forcelist=RETRY_STATUSES, raise_on_status=True)
        adapter = CustomSslContextHTTPAdapter(max_retries=retry)
        session.mount('https://', adapter)
        return session
    
    def call_data_source(self, query:UXDataSourceInput) -> 'UXDataSourceResponse':
        """
        Call the UX data source.

        Parameters:

        - query: UXDataSourceInput object

        Returns:

        - UXDataSourceResponse object
        """
        url = f'https://{self.url_db}cloud.plex.com/api/datasources/{query.__api_id__}/execute?format=2'
        session = self._create_session()
        response = session.post(url, json=query._query_string, auth=self._auth)
        json_data = response.json()
        return UXDataSourceResponse(query.__api_id__, **json_data)
    
    def list_data_source_access(self, pcn:HTTPBasicAuth|str|list):
        """
        Get a list of data sources that are enabled for a specific account or any number of accounts.

        Parameters:

        - pcn: Authentication for the account(s)

        Returns:

        - UXDataSourceResponse object
        """
        url = f'https://{self.url_db}cloud.plex.com/api/datasources/search?name='
        session = self._create_session()
        access_list = []
        if isinstance(pcn, list):
            pcn_list = pcn
        else:
            pcn_list = [pcn]
        for pcn in pcn_list:
            self._auth = self.set_auth(pcn)
            response = session.get(url, auth=self._auth)
            j = json.loads(response.text)
            for ds in j:
                ds['pcn'] = pcn
            access_list.append(j)
        all_datasources = list(chain.from_iterable(access_list))
        return UXDataSourceResponse('access_list', rows=all_datasources)

class UXDataSourceResponse(DataSourceResponse):
    def __init__(self, data_source_key, **kwargs):
        super().__init__(data_source_key, **kwargs)
        if isinstance(getattr(self, 'outputs', None), dict):
            for k, v in self.outputs.items():
                setattr(self, k, v)
        if getattr(self, 'rowLimitExceeded', False):
            warn('Row limit was exceeded in response. Review input filters and adjust to limit returned data.', category=UserWarning, stacklevel=3)
        if getattr(self, 'errors', []):
            raise UXResponseErrorLog(self.errors)
        self._format_response()

    def _format_response(self):
        self._transformed_data = getattr(self, 'rows', [])
        return self._transformed_data


from datetime import datetime
import os
import sys
import json
import csv
from warnings import warn
import logging
DEFAULT_FORMATTER = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
def debug_logger(level=logging.NOTSET):
    logger = logging.getLogger(__name__)
    FORMAT = "[%(asctime)s][%(filename)s:%(lineno)s][%(funcName)20s()] %(message)s"
    logging.basicConfig(format=FORMAT)
    logger.setLevel(level)
    return logger


def frozen_check():
    """
    Checks the running script to see if it is compiled to a single exe.
    If compiled, the resources will be stored in a temp folder.
    If not, then they will be in the script's working directory.
    """
    if getattr(sys, 'frozen', False):
    # Running in a bundle
        bundle_dir = sys._MEIPASS # pylint: disable=no-member
    else:
    # Running in a normal Python environment
        bundle_dir = os.path.dirname(os.path.abspath(__file__))
    return bundle_dir
def debug_dump_variables(obj):
    """
    Dumps variables of provided object to a log file.
    """
    if not hasattr(obj, 'dump_logger'):
        if hasattr(obj, 'batch_folder'):
            root = obj.batch_folder
        else:
            root = os.getcwd()
        obj.dump_logger = obj.setup_logger('Debug Dump', log_file='Debug_Dump', root_dir=root)
    obj.dump_logger.debug(f"Dumping variables for {type(obj.__name__)}:")
    for k, v in vars(obj).items():
        obj.debug_logger.debug(k, v)
        obj.dump_logger.debug(f'{k} : {v}')

def get_case_insensitive_key_value(input_dict, key):
    return next((value for dict_key, value in input_dict.items() if dict_key.lower() == key.lower()), None)

def create_batch_folder(root='', batch_code=None, include_time=False, test=False) -> str:
    """
    Used to set up a batch folder to store any log files or screenshots during an automation run.
    
    Parameters:

    - root: The root directory where the batch folder will be created.
    - batch_code: Optional batch code to use instead of generated one. Overrides include_time parameter.
    - include_time: If True, appends the current time to the batch code.
    - test: If True, uses 'TEST' for the batch folder path; otherwise, uses 'PROD'.

    Returns:
    
    - The path to the created batch folder.
    """
    if batch_code and include_time:
        warn('batch_code and time arguments are not supported together. Ignoring time argument.')
        include_time = False

    db = 'TEST' if test else 'PROD'
    now = datetime.now()
    b_code = batch_code or now.strftime('%Y%m%d')
    b_time = now.strftime('%H%M')
    folder_name = f'{b_code}_{b_time}' if include_time else b_code
    batch_folder = os.path.join(root, 'batch_codes', db, folder_name)
    os.makedirs(batch_folder, exist_ok=True)
    return batch_folder
def setup_logger(name, log_file='log.log', file_format='DAILY',
                     level=logging.DEBUG, formatter=DEFAULT_FORMATTER, root_dir=None) -> logging.Logger:
    """
    Setup a logging file.

    Parameters:

    - name: logger name
    - log_file: filename for the log file.
    - file_format: "DAILY" | "MONTHLY" | "". Will be combined with the log_file filename provided.
    - level: log level for the logger. logging module levels.
    - formatter: logging formatter
    - root_dir: root directory to store the log file

    Default formatter: %(asctime)s - %(name)s - %(levelname)s - %(message)s
    """
    _name = str(name)
    _file_format = str(file_format).upper()
    today = datetime.now()
    _formatter = logging.Formatter(formatter)
    if _file_format == 'DAILY':
        log_date = today.strftime("%Y_%m_%d_")
    elif _file_format == 'MONTHLY':
        log_date = today.strftime("%Y_%m_")
    else:
        log_date = ''
    _log_file = log_date + log_file
    if root_dir:
        _log_file = os.path.join(root_dir, _log_file)
    handler = logging.FileHandler(_log_file, mode='a', encoding='utf-8')
    handler.setFormatter(_formatter)
    logger = logging.getLogger(_name)
    logger.setLevel(level)
    if not logger.hasHandlers():
        logger.addHandler(handler)
    return logger

def read_updated(in_file):
    
    """
    Read in a file of already updated records.

    Parameters:
    
    - in_file: file containing the data to read.

    Returns:

    - json object or empty list
    """
    updated_records = []
    _file_type = in_file.split('.')[-1].lower()
    if os.path.exists(in_file):
        with open(in_file, 'r', encoding='utf-8') as f:
            if _file_type == 'json':
                updated_records = json.load(f)
            elif _file_type == 'csv':
                c = csv.DictReader(f)
                updated_records = [row for row in c]
            else:
                raise TypeError('File name provided is not an expected type of json or csv.')
    return updated_records or []

def save_updated(in_file, obj):
    """
    Save a file containing a list of already processed records.

    Parameters:
    
    - in_file: file to use to save
    - obj: json object to write to file. Expects a list containing dictionaries.
    """
    _file_type = in_file.split('.')[-1].lower()
    with open(in_file, 'w+', encoding='utf-8') as f:
        if _file_type == 'json':
            f.write(json.dumps(obj, indent=4))
        elif _file_type == 'csv':
            c = csv.DictWriter(f, fieldnames=obj[0].keys(), lineterminator='\n')
            c.writeheader()
            c.writerows(obj)
        else:
            raise TypeError('File name provided is not an expected type of json or csv.')


from pmc_automation_tools.api.common import (
    DataSourceInput,
    DataSourceResponse,
    DataSource,
    CustomSslContextHTTPAdapter,
    RETRY_COUNT,
    BACKOFF,
    RETRY_STATUSES
    )
from pmc_automation_tools.common.exceptions import ApiError
import requests
from requests.exceptions import HTTPError

from urllib3.util.retry import Retry

from itertools import chain

TEST = 'https://test.connect.plex.com'
PROD = 'https://connect.plex.com'

class ApiDataSourceInput(DataSourceInput):
    def __init__(self, url: str, method: str, *args, **kwargs):
        super().__init__(url, type='api', *args, **kwargs)
        self._method = method


    def _update_input_parameters(self):
        self._query_string = {k:v for k, v in vars(self).items() if not k.startswith('_')}


class ApiDataSource(DataSource):
    def __init__(self, auth: str, *args, test_db: bool = True, **kwargs):
        """
        Parameters:

        - auth: str
            - API Key as a string

        - test_db: bool, optional
            - Use test or production database
        """
        super().__init__(*args, auth=auth, test_db=test_db, type='api', **kwargs)
    
    
    def call_data_source(self, pcn:str|list, query:ApiDataSourceInput):
        """
        Returns a list of the json objects as dictionaries from the API response.

        Parameters:

        - pcn: str | list
            - Single PCN number or list of PCNs to run the query against

        - query: ApiDataSourceInput
            - DataSourceInput containing the connection parameters
        """
        if self._test_db:
            query.__api_id__ = query.__api_id__.replace(PROD, TEST)
        response_list = []
        if isinstance(pcn, str):
            pcn_list = [pcn]
        for p in pcn_list:
            headers = {'Content-Type': 'application/json',
                'X-Plex-Connect-Api-Key': self._auth,
                'X-Plex-Connect-Customer-Id': p
            }
            session = requests.Session()
            retry = Retry(total=RETRY_COUNT, connect=RETRY_COUNT, backoff_factor=BACKOFF, status_forcelist=RETRY_STATUSES, raise_on_status=True)
            adapter = CustomSslContextHTTPAdapter(max_retries=retry)
            session.mount('https://', adapter)
            request_params = {'json': query._query_string} if query._method.upper() in ['POST', 'PUT'] else {'params': query._query_string}
            response = session.request(query._method, query.__api_id__, headers=headers, **request_params)
            try:
                response.raise_for_status()
            except HTTPError as e:
                raise ApiError('Error calling API.', **response.json(), status=response.status_code)
            # TODO - Check how the various response schema should be handled
            if response.status_code == 200 and response.text != [] and response.text != '':
                response_list.append(response.json())
            else:
                return response
        return ApiDataSourceResponse(query.__api_id__, response_list = list(chain.from_iterable(response_list)))
    

class ApiDataSourceResponse(DataSourceResponse):
    def __init__(self, url, **kwargs):
        super().__init__(url, **kwargs)
        self._format_response()

    def _format_response(self):
        self._transformed_data = getattr(self, 'response_list', [])
